"""Module containing functionality for loading projects from unpacked local sources."""

import contextlib
import shutil
import tempfile
from collections.abc import Generator
from dataclasses import dataclass
from pathlib import Path
from typing import Any

from cargo2rpm.metadata import Metadata, Package
from cargo2rpm.semver import Version

from rust2rpm import log
from rust2rpm.cli import Options
from rust2rpm.conf import TomlConf
from rust2rpm.inspect import get_doc_files, get_license_files
from rust2rpm.metadata import guess_main_package
from rust2rpm.patching import make_patches
from rust2rpm.vendor import generate_vendor_tarball

from .meta import PatchFile, Project
from .shared import file_name_auto_patch, file_name_manual_patch


@contextlib.contextmanager
def temp_project_copy(path: Path) -> Generator[Path, Any, None]:
    """Provide a copy of the project sources in a temporary directory."""
    with tempfile.TemporaryDirectory() as tmpdir:
        target = Path(tmpdir) / path.name
        yield shutil.copytree(path, target)


def guess_local_project_version_from_dir(dir_name: str) -> tuple[str | None, str | None]:
    """Guess name and version of a local project based on the name of the parent directory.

    The heuristic used in this function does not work for pre-release versions,
    since a default format of `{name}-{version}` is assumed, but pre-releases
    contain a `-` character too.

    Arguments:
        dir_name: Name of the directory.

    Returns:
        Tuple of (name, version) strings, or (`None`, `None`) in case the
            directory name could not be parsed.

    """
    project = dir_name.rstrip("0123456789.").removesuffix("-")
    version = dir_name.removeprefix(f"{project}-")

    try:
        Version.parse(version)
    except ValueError:
        return None, None
    else:
        return project, version


def guess_local_project_version_from_path(path: Path, version: str | None) -> tuple[str | None, str | None]:
    """Guess name and version of a local project based on the project path and version.

    If the version is specified, it is stripped from the name of the parent
    directory, and the remaining string is assumed to the the project name.

    If the version is not specified, the name of the parent directory is
    attempted to be parsed based on the assumption that it is named according to
    the common `{name}-{version}` format.

    Arguments:
        path: Path to the project's root directory or to the Cargo.toml file at
            the project root.
        version: Optional version string. This is useful for pre-releases (which
            otherwise cannot be parsed correctly) or if the upstream project
            releases tarballs with non-SemVer version strings.

    Returns:
        Tuple of (name, version) strings, or (`None`, `None`) in case the
            directory name could not be parsed.

    """
    dir_name = path.resolve().name if path.is_dir() else path.resolve().parent.name

    if version:
        project = dir_name.removesuffix(f"-{version}")
        return project, version

    return guess_local_project_version_from_dir(dir_name)


@dataclass(frozen=True)
class LocalProject(Project):
    """Description of a project based on locally unpacked sources."""

    main_package: Package
    """Heuristically determined "main" package of the cargo workspace.

    In the case of local sources that contain a single crate (i.e. which are not
    cargo workspaces), this contains the metadata of this single crate.
    """

    @property
    def is_workspace(self) -> bool:
        """Predicate that is True for cargo workspaces and False for single crates."""
        return len(self.metadata.packages) > 1

    @property
    def rpm_name(self) -> str:
        """RPM source package name based on project metadata."""
        return self.name

    @staticmethod
    def from_cargo_toml_path(
        path: Path,
        name: str | None,
        version: str | None,
        out_dir: Path | None,
        conf: TomlConf,
        options: Options,
    ) -> "LocalProject":
        """Load project metadata based on the contents of an unpacked source tree.

        Arguments:
            path: Path to the Cargo.toml file at the root of the project.
            name: Optional name hint for the project in case it cannot be
                guessed from the name of the parent directory.
            version: Optional version hint for the project in case it cannot be
                guessed from the name of the parent directory.
            out_dir: Output directory for any generated or downloaded files.
            conf: Global rust2rpm configuration.
            options: Options and flags derived from command-line arguments.

        Returns:
            Project metadata based on the contents of the unpacked source tree.

        """
        project_root = path.parent
        doc_files = get_doc_files(project_root, conf)
        license_files = get_license_files(project_root, conf)

        metadata = Metadata.from_cargo(str(path))
        parent_dir = Path(project_root).parent

        is_workspace = len(metadata.packages) > 1

        if is_workspace:
            log.info("Skipping automatic creation of patches for cargo workspace.")

            # fall back to the directory name for determining the name / version
            # of the project heuristically
            guessed_name, guessed_version = guess_local_project_version_from_path(path, version)

            main_package = guess_main_package(metadata, name, guessed_name)

            # prefer explicit arguments over guesses over conjectures
            name = name or guessed_name or main_package.name
            version = version or guessed_version or main_package.version

            log.warn(
                f"Falling back to {name!r} as the main package. If this is not correct, "
                "specify the correct name in the 'pkgid' argument on the command line.",
            )

            vendor_tarball = generate_vendor_tarball(path, name, version, out_dir or parent_dir, options.vendor)

            return LocalProject(
                name,
                version,
                metadata,
                license_files,
                doc_files,
                None,
                None,
                vendor_tarball,
                main_package,
            )

        # RET505: removing this else branch makes this less readable
        else:  # noqa: RET505
            package = metadata.packages[0]

            name = package.name
            version = package.version

            reapply_path = (out_dir or Path()) / file_name_manual_patch(name) if options.reuse_patch else None

            with temp_project_copy(project_root.absolute()) as temp_root:
                temp_toml = temp_root / "Cargo.toml"

                auto_diff, manual_diff = make_patches(
                    name,
                    package.version,
                    temp_toml,
                    reapply_path,
                    options,
                )

                # ensure metadata is up-to-date with changes from patches
                metadata = Metadata.from_cargo(str(temp_toml))

                vendor_tarball = generate_vendor_tarball(
                    temp_toml,
                    name,
                    version,
                    out_dir or parent_dir,
                    options.vendor,
                )

            auto_patch = PatchFile(file_name_auto_patch(name), diff) if (diff := auto_diff) else None
            manual_patch = PatchFile(file_name_manual_patch(name), diff) if (diff := manual_diff) else None

            return LocalProject(
                name,
                version,
                metadata,
                license_files,
                doc_files,
                auto_patch,
                manual_patch,
                vendor_tarball,
                package,
            )

    @staticmethod
    def from_project_folder(
        path: Path,
        name: str | None,
        version: str | None,
        out_dir: Path | None,
        conf: TomlConf,
        options: Options,
    ) -> "LocalProject":
        """Load project metadata based on the contents of an unpacked source tree.

        This method is a thin wrapper around `LocalProject.from_cargo_toml_path`.

        Arguments:
            path: Path to the root directory of the unpacked sources.
            name: Optional name hint for the project in case it cannot be
                guessed from the name of the parent directory.
            version: Optional version hint for the project in case it cannot be
                guessed from the name of the parent directory.
            out_dir: Output directory for any generated or downloaded files.
            conf: Global rust2rpm configuration.
            options: Options and flags derived from command-line arguments.

        Returns:
            Project metadata based on the contents of the unpacked source tree.

        """
        toml_path = path / "Cargo.toml"

        return LocalProject.from_cargo_toml_path(
            toml_path,
            name,
            version,
            out_dir,
            conf,
            options,
        )
